

program MBEx; {entwickelt mit FPC 3.0.4 fr Windows i386/x86_64}


  {$APPTYPE           GUI} {ohne DOS-Fenster beim Programmstart        }
  {$DEBUGINFO         OFF} {keine Debug-Informationen generieren       }
  {$EXTENDEDSYNTAX     ON} {erweiterte Syntax, z.B. fr API-Aufrufe    }
  {$MINFPCONSTPREC     32} {Konstanten mglichst als "single"          }
  {$MODE              FPC} {Kompatibilitts-Modus fr Free Pascal      }
  {$MODESWITCH EXCEPTIONS} {"try" erlaubt, hier fr "FPWritePNG"       }
  {$POINTERMATH        ON} {Pointer-Arithmetik erlauben                }
  {$PROFILE           OFF} {keinen "Profiling Code" generieren         }
  {$S-                   } {Stack nicht berwachen                     }
  {$SMARTLINK          ON} {nur verwendete Prozeduren mit einbinden    }

  {$ALIGN              ON} {"record"-Daten ausrichten                  }
  {$ASMMODE         INTEL} {Assembler in Intel-Syntax                  }
  {$ASSERTIONS        OFF} {kein "Assertion Support"                   }
  {$BOOLEVAL          OFF} {Boolean-Ausdrcke nicht komplett auswerten }
  {$CHECKPOINTER      OFF} {Pointer und Heap nicht berwachen          }
  {$COPERATORS        OFF} {keine C-hnliche Syntax                    }
  {$GOTO               ON} {"goto" und Labels erlauben                 }
  {$LONGSTRINGS       OFF} {Shortstrings verwenden, entspricht $H-     }
  {$IOCHECKS          OFF} {kein I/O-Checking, entspricht $I-          }
  {$MACRO              ON} {Macros erlauben                            }
  {$MMX               OFF} {FPU-Nutzung statt MMX-Support              }
  {$OPTIMIZATION       ON} {Optimierung durch den Compiler             }
  {$OVERFLOWCHECKS    OFF} {kein Overflow-Checking fr Integer         }
  {$RANGECHECKS       OFF} {kein Range-Checking                        }
  {$SAFEFPUEXCEPTIONS OFF} {kein "fwait" nach FPU-Befehlen             }
  {$STACKFRAMES       OFF} {kein "Stack Frame" fr einfache Prozeduren }
  {$TYPEDADDRESS      OFF} {"@"-Pointer kompatibel zu allen Pointern   }
  {$TYPEINFO          OFF} {keine "Run-Time Type Information"          }
  {$VARSTRINGCHECKS   OFF} {Compiler prft var-Strings nicht           }
  {$WARNINGS           ON} {alle Compiler-Warnungen anzeigen           }


  uses SysUtils, DOS, ZStream, FPImage, FPWritePNG, Windows;


  {$if     defined (WIN32) and defined (CPUI386)   and (SizeOf (extended) = 10)} {$define x32}
  {$elseif defined (WIN64) and defined (CPUX86_64) and (SizeOf (extended) =  8)} {$define x64}
  {$else} {$fatal +++ not FPC 3.0.4 for Windows i386/x86_64 +++} {$endif}

  {$if (FPC_FULLVERSION <> 30004)} {$warning +++ not FPC version 3.0.4 +++} {$endif}

  {$ifdef x32} {$info x32} {$FPUTYPE X87} {voreingestellt ist "-CpPENTIUM"}

    type  uP_real = extended;
    type  uP_word = longword;
    type  uP_int  = longint;

  {$else} {$info x64} {$FPUTYPE SSE64} {voreingestellt bzw. kein Unterschied zwischen "-CpATHLON64" und "-CpCOREI"}

    type  uP_real = double;
    type  uP_word = qword;
    type  uP_int  = int64;

  {$endif}


  {$resource APPICON.RES}
  {$include  Font.hdr}


  type long128 = array [0 ..  3] of longword; {128 Bit fr alle "j"-Variablen}
  type quad128 = record L, H : qword end;
  type word128 = array [0 ..  7] of word;
  type byte128 = array [0 .. 15] of byte;

  type word80  = array [0 ..  4] of word; {80 Bit fr das "extended"-Format}
  type byte80  = array [0 ..  9] of byte;

  type word64  = array [0 ..  3] of word; {64 Bit fr das "double"-Format}
  type byte64  = array [0 ..  7] of byte;

  type byte32  = array [0 .. 3] of byte; {typecast fr "longword" und "word"}
  type byte16  = record L, H : byte end;


  const Spektrum_Tab : array [0 .. 3, 0 .. 7] of byte =  {  0    1    2    3    4    5    6    7}
                                                        ((  0,  27, 114, 157, 226, 253, 254, 255),  {Sttzstellen im "Spektrum"}
                                                         {--------------------------------------}
                                                         (  0,  34, 237, 255, 128,   0,   0,   0),  {R}
                                                         (  0,  47, 255, 165,   0,   0,   0,   0),  {G}
                                                         (  0, 174, 255,   0,   0,   0,   0,   0)); {B}


  const HD_X = 1920; HD_Y = 1200; HD_Size = (HD_X * HD_Y); {HD bzw. WUXGA als maximale Auflsung}

  const N_X = (HD_X + 128); N_Y = (HD_Y + 128); N_Size = (N_X * HD_Y); N_real_Size = (N_X * N_Y); {"N_array" mit 64 Pixel Rand}

  var PNG_Image : TFPMemoryImage; PNG_Writer : TFPCustomImageWriter; {fr die Units "FPImage" und "FPWritePNG"}            {ca. 18 MB Speicher}

  var Buf_N_array       : array [0 .. 3, 0 .. (N_Size - 1)] of longword; {Buffer fr die letzten Bilder}                   {ca. 40 MB Speicher}
  var Buf_jCenter_zX    : array [0 .. 3] of  long128;
  var Buf_jCenter_zY    : array [0 .. 3] of  long128;
  var Buf_ld_Zoom       : array [0 .. 3] of  longint;
  var Buf_Iteration     : array [0 .. 3] of longword;
  var Buf_Color         : array [0 .. 3] of   double;
  var Buf_Offset        : array [0 .. 3] of longword;
  var Buf_Alt_Iteration : array [0 .. 3] of longword;
  var Buf_Alt_Color     : array [0 .. 3] of   double;
  var Buf_Automatik     : array [0 .. 3] of  boolean;
  var Buf_FPU_Mode      : array [0 .. 3] of  boolean;
  var Buf_CPU_Prev      : array [0 .. 3] of  longint;  {... und ...}  Buf_Pointer : longword;

  var Spektrum : array [0 .. 255] of longword; {wandelt N aus "N_array" in RGB-Pixels}                                     {nur  1 kB Speicher}

  var Cursor, Pixel : array [0 .. (HD_Size - 1)] of longword; {Array mit den RGB-Pixels als "Device Independent Bitmap"}   {ca. 18 MB Speicher}

  var N_Anf   : array [0 .. (N_X *   64) + 64 - 1] of longword; {"N_array" merkt die Anzahl der Iterations-Schritte}       {ca. 23 MB Speicher}
  var N_array : array [0 .. (N_X * HD_Y)      - 1] of longword;
  var N_End   : array [0 .. (N_X *   64) - 64 - 1] of longword;
  var N_Ctrl  : array [0 .. (N_X * HD_Y)      - 1] of longword; {steuert die Iteration in den Threads und ...   }
  var N_Check : array [0 .. (N_X * HD_Y)      - 1] of  boolean; {... die Rekursion bei eingeschalteter Automatik}

  const Work = $FFFFFFFF; {Maske fr "N_Ctrl"}                                                                    {in Summe ca. 99 MB Speicher}

  const App_Name = 'MBEx'; var Window : hWnd; WindowDC : hDC; {API-Core}
  var wcls : tWndClassEx; {fr "RegisterClassEx" - siehe "FPC-Source\WinAPI\WinAPI-examples\BitBlt\BitBlt.pas"}
  var Red_Brush, Green_Brush, Dark_Brush, Gray_Brush, Light_Brush : hBrush; {zum schnellen Zeichnen mit "Area"}
  var Area, Area_L, Area_O, Area_R, Area_U, Area_z : tRect;
  var dib : BitmapInfo; {"Device Independent Bitmap" - siehe "FPC-Source\fpc-3.0.4\rtl\win\wininc\struct.inc"}

  var ParamStr_Threads, ParamStr_Window : longint; ParamStr_TaskBar : boolean; {Parameter von "ParamStr" und Parameter-File}
  var ParamStr_FPU, Param_loaded : shortint; Picture_Path, File_Name : string;
  var Param_jCenter_zX, Param_jCenter_zY : long128; Param_ld_Zoom : longint;
  var Param_Color, Param_Alt_Color : double; Param_Offset, Param_Iteration, Param_Alt_Iteration : longword;
  var Param_Automatik, Param_FPU_Mode : boolean; Param_CPU_Prev : longint;

  var Hard_X, Hard_Y, Screen_X, Screen_Y, DPI_Scale, Window_X, Window_Y, Rand_L, Rand_O : longint; {Bildschirm und Rahmen}
  var Fullscreen, Red_Frame : boolean;

  var Wait_for_first_Paint, Wait_for_Init_Buf : boolean; {Abblauf-Steuerung}
  var Double_Key : longword; Mouse_pressed, Mouse_Timer, Info_Screen, Message_Timer : boolean; Message : shortint;
  var Focus_lost, Focus_Restore_locked, Focus_Leave_Request : boolean; XP_Try_Restore : longint;

  type Field = record L, R, O, U : longword end; {Buttons fr die Maus}
  var Field_Info, Field_Exit, Field_Back, Field_Reset, Field_Zoom_x16, Field_Zoom_x2, Field_Zoom_out, Field_Color : Field;
  var Field_Info_F1, Field_Info_Auto, Field_Info_More, Field_Info_Less, Field_Info_Prev, Field_Info_Late : Field;
  var Field_Info_F7, Field_Info_F8, Field_Info_Save, Field_Info_F10, Field_Info_F11, Field_Info_F12, Field_Message : Field;

  var Mouse_X, Mouse_Y : longword; Move_X, Move_Y : longint; Move_happens : boolean; Mouse_Color : double; {Maus und "Move"}
  var dst_L, dst_O, Move_B, Move_H, Move_Black : longword;

  var Kachel, Call_Kachel, Break_Cnt, Break_Anz : longint; Rekursion, Estimate : longword; {steuert die "Paint_Action" und die Threads}
  var Kachel_Ausgabe, Break_Flag, Automatik, Rekursion_OFF : boolean; const NOP = 777;

  var Start_zX, Start_zY, Delta_z : uP_real; jCenter_zX, jCenter_zY, jStart_zX, jStart_zY, jDelta_z : long128;
  var iStart_zX, iStart_zY : uP_int; iDelta_z : uP_word; ld_Zoom : longint; Color, Alt_Color : double; Offset : longword;
  var Iteration, Alt_Iteration : longword; Out_of_Range, FPU_Mode : boolean; CPU_Mode : byte; CPU_Prev : longint;

  const Max_Iteration = round (1.0E+07 - 1.0); {Maximal-Wert fr die Iteration - siehe "Decide_Rekursion_with_Estimate" im Paint-Header}

  var QPC_pro_s, Break_QPC, Idle_QPC, t_Start, t_last, Laufzeit : int64; {Zeitmessung}

  var Anzahl_Threads : longint; Thread_ID : array [1 .. 8] of TThreadID; {krzere Laufzeit, wenn die Threads erst am Schlu kommen}
  var Thread_Start, Thread_Ready : array [1 .. 8] of PRTLEvent;
  var Thread_L, Thread_R, Thread_O, Thread_U : array [1 .. 8, 0 .. 1023] of longint;
  var  Delta_O : uP_real;  Delta_L : array [1 .. 8] of uP_real;
  var iStart_O : uP_int ; iStart_L : array [1 .. 8] of uP_int ;
  var jStart_O : long128; jStart_L : array [1 .. 8] of long128;

  {$ifdef x32} var FPU_Limit : array [1 .. 8] of longint; FPU_last64, FPU_Bits : longint; FPU_Control_Word : word = $037F; {$endif}
  {$ifdef x32} const CPU_last28 = 16; {max. 20} {$else} const CPU_last60 = 52; {max. 52} {$endif}
  const FPU_last53 = 45; {max. 45}


  {--- wichtige Prozeduren ---}   {-----------------------------------------------------------------------------------------------------------}


  procedure Abbruch (S : string); begin
    S := (S + #0); MessageBox (0, @S [1], 'ERROR' + #0, mb_OK or mb_IconStop or mb_TopMost or mb_SetForeground); Halt (1);
  end;

  function t : int64; var QPC : int64; begin QueryPerformanceCounter (QPC); t := round ((QPC * 1000.0) / QPC_pro_s) end; {Zeit in ms}

  procedure Trigger_Paint_Action; begin InvalidateRect (Window, NIL, true) end; {lst "wm_Paint" aus}

  procedure Kachel_Job (I : longint); begin Kachel := I; if (GetQueueStatus (qs_Paint) = 0) then Trigger_Paint_Action end;

  procedure Set_N_Ctrl (x : longword); begin FillDWord (N_Ctrl, N_Size, x) end; {"N_Ctrl = Work" oder "N_Ctrl = 0"}

  procedure Set_N_Check (x : boolean); begin FillByte (N_Check, N_Size, byte (x)) end; {"N_Check = true" oder "N_Check = false"}

  {$define Spektrum_from_N := Spektrum [(round (longint (Iteration - N) * Color) + Offset) and $FF] }  {kommt im Paint- und Thread-Header vor}


  {$include MBEx-long128.hdr}
  {$include MBEx-Threads.hdr}


  {$ifdef x32} function New_Mode : byte; begin {neuen "CPU_Mode" fr x32 ermitteln}

    if Out_of_Range then begin New_Mode := 1; {+++} exit {+++} end;

    if ld_Zoom > (FPU_last64 - CPU_Prev) then begin New_Mode := 3; {+++} exit {+++} end;
    if ld_Zoom > (FPU_last53 - CPU_Prev) then begin New_Mode := 2; {+++} exit {+++} end;
    if ld_Zoom > (CPU_last28 - CPU_Prev) then begin New_Mode := 1; {+++} exit {+++} end;

    if FPU_Mode then New_Mode := 1 else New_Mode := 0;

  end; {endfunc} {$endif}


  {$ifdef x64} function New_Mode : byte; begin {neuen "CPU_Mode" fr x64 ermitteln}

    if Out_of_Range then begin New_Mode := 1; {+++} exit {+++} end;

    if ld_Zoom > (CPU_last60 - CPU_Prev) then begin New_Mode := 3; {+++} exit {+++} end;
    if ld_Zoom > (FPU_last53 - CPU_Prev) then begin New_Mode := 0; {+++} exit {+++} end;

    if FPU_Mode then New_Mode := 1 else New_Mode := 0;

  end; {endfunc} {$endif}


  procedure Set_Variables_SYS; var dX, dY, Ende_zX, Ende_zY : uP_real; QPC : int64; begin {wird von allen "SYS"-Prozeduren aufgerufen}

  {$ifdef x32}

    word80 (Delta_z) [4] := ($3FF7 - ld_Zoom); {synthetisches "extended" verringert den Quantisierungs-Fehler fr alle Formate, ...   }
    word80 (Delta_z) [3] := $8000;             {... auerdem mu "jCenter" dann auch mit kurzen 2er-Potenzen initialisiert worden sein}
    word80 (Delta_z) [2] := 0;
    word80 (Delta_z) [1] := 0;
    word80 (Delta_z) [0] := 0; {exakt 3.90625E-03 fr "ld_Zoom = 0"}

  {$else} {x64}

    word64 (Delta_z) [3] := $3F70 - (ld_Zoom shl 4); {synthetisches "double"}
    word64 (Delta_z) [2] := 0;
    word64 (Delta_z) [1] := 0;
    word64 (Delta_z) [0] := 0; {auch hier exakt 3.90625E-03 fr "ld_Zoom = 0"}

  {$endif}

     jDelta_z := jMake ( Delta_z); {im Folgenden ist die Referenz neben "Delta_z" immer "jCenter" mit 120 Bit Genauigkeit}
     iDelta_z := iMake (jDelta_z);
           dX := ((Window_X * Delta_z) / 2.0);
           dY := ((Window_Y * Delta_z) / 2.0);
    jStart_zX := jSub (jCenter_zX, jMake (dX)); {z-Koordinate links oben im "Window" - Referenz ist "jCenter"}
    jStart_zY := jSub (jCenter_zY, jMake (dY));
     Start_zX := jReload (jStart_zX);
     Start_zY := jReload (jStart_zY);
    iStart_zX := iMake   (jStart_zX);
    iStart_zY := iMake   (jStart_zY);
      Ende_zX := jReload (jCenter_zX) + dX; {z-Koordinate rechts unten, wird nur zum berprfen gebraucht - s.u.}
      Ende_zY := jReload (jCenter_zY) + dY;

    if (Color = Alt_Color) then begin {Farbe anpassen}
      Color := (4.0 - ((3.0 / 50.0) * ld_Zoom)); {Wertebereich [4 .. 1] fr "ld_Zoom" aus [0 .. 50]}
      if (Color < 1.0) then Color := 1.0;
      Alt_Color := Color;
    end else begin
      Alt_Color := 0.0; {Farb-Anpassung abschalten}
    end;

    Out_of_Range := (abs (Start_zX) >= 3.9) or (abs (Start_zY) >= 3.9) or                 {weder i- noch j-Mode ...}
                    (abs ( Ende_zX) >= 3.9) or (abs ( Ende_zY) >= 3.9) or (ld_Zoom < 0);  {... erlauben groe Werte}

    CPU_Mode := New_Mode; {neuen "CPU_Mode" ermitteln}

    t_Start := t; {Zeitmessung starten}                             {+++ wichtig fr die "Paint_Action", deswegen auch die "SYS"-Namen +++}
    Kachel_Ausgabe := true; {"Kachel_Ausgabe" erlauben}
    Rekursion := 0; {laufende Rekursion auf jeden Fall abbrechen}
    QueryPerformanceCounter (QPC); Idle_QPC := (QPC + QPC_pro_s); {Limit fr KEINE "Kachel_Ausgabe"}

  end; {endproc}


  {$include MBEx-File.hdr}
  {$include MBEx-Info.hdr}
  {$include MBEx-Paint.hdr}


  {--- Buffer-Verwaltung ---}   {-------------------------------------------------------------------------------------------------------------}


  procedure Write_Buf; label NEXT, STAY; begin {wird aufgrufen, wenn ein Zyklus erfolgreich beendet wurde}

    if  Wait_for_Init_Buf  then goto STAY; {fr "Init_Buf"     }

    if (Call_Kachel = -10) then goto STAY; {Bild neu berechnen } {aktuellen Buffer berschreiben}
       {               -9}                 {Farbe ndern o..  }
    if (Call_Kachel =  -8) then goto NEXT; {16-facher Zoom     } {alle Zoom-Funktionen stellen den Buffer weiter}
    if (Call_Kachel =  -7) then goto NEXT; {2-facher Zoom      }
    if (Call_Kachel =  -6) then goto NEXT; {Zoom zurck        }        {--------------------------------------------------}
    if (Call_Kachel =  -5) then goto STAY; {mehr Iterationen   }        { z.T. nur zur Dokumentation, denn wenn der Zyklus }
       {               -4}                 {weniger Iterationen}        { nicht gestartet wird (z.B. "Call_Kachel = -9"),  }
       {               -3}                 {"Mouse-Move"       }        { entfllt mit den Zyklus-Ende auch "Write_Buf"    }
    if (Call_Kachel =  -2) then goto STAY; {Bild verschieben   }        {--------------------------------------------------}
    if (Call_Kachel =  -1) then goto NEXT; {System-Reset       }
    {+++} exit {+++};                      {alles andere       }

    NEXT: Buf_Pointer := (Buf_Pointer + 1) and $03; STAY: {abhngig von "Call_Kachel" Buffer weiterstellen oder berschreiben}

    Buf_N_array       [Buf_Pointer] := N_array;
    Buf_jCenter_zX    [Buf_Pointer] := jCenter_zX;
    Buf_jCenter_zY    [Buf_Pointer] := jCenter_zY;
    Buf_ld_Zoom       [Buf_Pointer] := ld_Zoom;
    Buf_Iteration     [Buf_Pointer] := Iteration;
    Buf_Color         [Buf_Pointer] := Color;
    Buf_Offset        [Buf_Pointer] := Offset;
    Buf_Alt_Iteration [Buf_Pointer] := Alt_Iteration;
    Buf_Alt_Color     [Buf_Pointer] := Alt_Color;
    Buf_Automatik     [Buf_Pointer] := Automatik;
    Buf_FPU_Mode      [Buf_Pointer] := FPU_Mode;
    Buf_CPU_Prev      [Buf_Pointer] := CPU_Prev;

  end; {endproc}


  procedure Read_Buf_Kachel_Job; label STAY; begin
                                                                                                {kann der Benutzer nach Programm-Start mit ...}
    if Wait_for_Init_Buf then begin Param_loaded := 0; Kachel_Job (-1); {+++} exit {+++} end;   {... zeitaufwendigem Parameter-File auslsen  }

    if (Kachel <> 9999) then goto STAY; {laufender Zyklus abgebrochen - aktuellen Buffer zurckladen}

    if (Buf_Iteration [Buf_Pointer] > Iteration) and
       (Buf_ld_Zoom   [Buf_Pointer] = ld_Zoom  ) then goto STAY; {Iteration verringert - aktuellen Buffer zurckladen}

    Buf_Pointer := (Buf_Pointer - 1) and $03; STAY: {entweder letzten oder aktuellen Buffer zurckladen}

    N_array       := Buf_N_array       [Buf_Pointer];
    jCenter_zX    := Buf_jCenter_zX    [Buf_Pointer];
    jCenter_zY    := Buf_jCenter_zY    [Buf_Pointer];
    ld_Zoom       := Buf_ld_Zoom       [Buf_Pointer];
    Iteration     := Buf_Iteration     [Buf_Pointer];
    Color         := Buf_Color         [Buf_Pointer];
    Offset        := Buf_Offset        [Buf_Pointer];
    Alt_Iteration := Buf_Alt_Iteration [Buf_Pointer];
    Alt_Color     := Buf_Alt_Color     [Buf_Pointer];
    Automatik     := Buf_Automatik     [Buf_Pointer];
    FPU_Mode      := Buf_FPU_Mode      [Buf_Pointer];
    CPU_Prev      := Buf_CPU_Prev      [Buf_Pointer];

    Laufzeit := 0; Estimate := NOP; Kachel_Job (-9); {es hat keine Berechnung stattgefunden - fr "Show_Info"}

  end; {endproc}


  procedure Init_Buf; var I : longword; begin
    for I := 0 to 3 do begin Buf_Pointer := I; {+++} Write_Buf {+++} end; Wait_for_Init_Buf := false;
  end;


  {--- "Paint_Action" ---}   {----------------------------------------------------------------------------------------------------------------}


  procedure User_Create; forward;
  procedure Hide_Window; forward;


    //  "Hide_Window" im API-Header
    //   setzt "not Mouse_pressed", "XP_Try_Restore = -1", "Message = 0" und "Focus_lost" (u.a. zur Maus- und Timer-Kontrolle).
    //  "Show_Message" im Info-Header
    //   setzt "Param_loaded = 0", "Message = -1" und "Message_Timer" (der "Message_Timer" kommt nach der "Paint_Action").


  function Color_OK : boolean; begin Color_OK := true;
    if (Color >= 1.1000e+09) then begin Color_OK := false; Color := 1.0001e+09; Message := 91 end;
    if (Color <= 0.9091e-09) then begin Color_OK := false; Color := 0.9999e-09; Message := 92 end;
  end;

  function Max_Zoom_OK : boolean; begin Max_Zoom_OK := true;
    if (ld_Zoom >= 112) then begin Max_Zoom_OK := false; Message := 87 end;
  end;

  function Min_Zoom_OK : boolean; begin Min_Zoom_OK := true;
    if (ld_Zoom <= -4) then begin Min_Zoom_OK := false; Message := 6 end;
  end;

  function Max_Iteration_OK : boolean; begin Max_Iteration_OK := true;
    if (Iteration >= Max_Iteration) then begin Max_Iteration_OK := false; Message := 5 end;
  end;

  function Min_Iteration_OK : boolean; begin Min_Iteration_OK := true;
    if (Iteration <= 2) then begin Min_Iteration_OK := false; Message := 4 end;
  end;

  procedure Override_by_ParamStr_FPU; begin
    if (ParamStr_FPU = -2) then begin FPU_Mode := false; CPU_Prev := 4 end;
    if (ParamStr_FPU =  0) and (Anzahl_Threads > 4) then FPU_Mode := false; {"Anzahl_Threads" entscheidet ber "FPU_Mode"}
  end;


  procedure Paint_Action; var L, O, U, B, H : longint; QPC : int64; label NEXT, ZYKLUSENDE, KACHEL_9999, READY; begin

    if (Kachel >= 1024) then {+++} exit {+++}; {d.h. auch bei "Kachel = 9999" - das kommt vor, weil die API oft sinnlose Aufrufe generiert}

    if (Kachel < 0) then begin {"Kachel_Job"}                                                     {-------------------------------------------}
                                                                                                  { Bei "Kachel = 9999" wird der Zyklus nicht }
      Call_Kachel := Kachel; {fr "Write_Buf" am Zyklus-Ende merken}                              { gestartet, ohne Zyklus-Ende entfllt dann }
                                                                                                  { die Laufzeit-Berechnung samt "Write_Buf". }
      if (Message < 0) then Message := 0; {die Message-Anzeige wird jetzt sicher gelscht}        {-------------------------------------------}

      if (Kachel = -10) then begin Make_N_array_SYS; Kachel := 0 end;                                                      {Bild neu berechnen}

      if (Kachel = -9) then if (Message = 0) then if Color_OK then begin    {u.a. Farbe ndern - Zyklus NICHT starten, aber "Buf" immer merken}
        Buf_Color     [Buf_Pointer] :=     Color;
        Buf_Offset    [Buf_Pointer] :=    Offset;
        Buf_Alt_Color [Buf_Pointer] := Alt_Color;
        Buf_FPU_Mode  [Buf_Pointer] :=  FPU_Mode;
        Buf_CPU_Prev  [Buf_Pointer] :=  CPU_Prev; Show_N_array_SYS; Kachel := 9999;
      end;

      if (Kachel = -8) then if Max_Zoom_OK then begin Make_Zoom_in_SYS (16, 4); Kachel := 0 end;                               {16-facher Zoom}

      if (Kachel = -7) then if Max_Zoom_OK then begin Make_Zoom_in_SYS (2, 1); Kachel := 0 end;                                 {2-facher Zoom}

      if (Kachel = -6) then if Min_Zoom_OK then begin Make_Zoom_out_SYS; Kachel := 0 end;                                            {Zoom out}

      if (Kachel = -5) then if Max_Iteration_OK then begin Make_more_Iteration_SYS; Kachel := 0 end;                         {mehr Iterationen}

      if (Kachel = -4) then if Min_Iteration_OK then begin                                                           {weniger Iterationen, ...}
        {+++} Laufzeit := 0; Estimate := NOP {+++}; Make_less_Iteration; Show_N_array_SYS; Kachel := 9999;           {... Zyklus NICHT starten}
      end;

      if (Kachel = -3) then begin Move_Picture_on_Screen; Kachel := 9999 end;                             {"Mouse-Move" - Zyklus NICHT starten}

      if (Kachel = -2) then begin Move_Picture_on_Screen; Move_Picture_in_Memory_SYS; Kachel := 0 end;           {Bild im Speicher verschieben}

      if (Kachel = -1) then begin                                                                            {System-Reset oder Programm-Start}
        {|}
        {|}  if Wait_for_first_Paint then begin
        {|}    {+++} User_Create {+++}; Wait_for_first_Paint := false; {$ifdef x32} Start_FPU_Test_Threads; {$endif}
        {|}  end;
        {|}
        {|}  if (Param_loaded <> 1) then begin
        {|}    jCenter_zX := jMake (0.0)     ;  jCenter_zY    := jMake (-0.5)       ;  {"-0.5" macht kaum Quantisierungs-Fehler}
        {|}    ld_Zoom    := 0               ;
        {|}    Iteration  := 256             ;  Alt_Iteration := 256                ;
        {|}    Color      := 4.0             ;  Alt_Color     := 4.0                ;  Offset   := 0;
        {|}    Automatik  := true            ;  FPU_Mode      := true               ;  CPU_Prev := 0;  {+++} Override_by_ParamStr_FPU {+++};
        {|}  end else begin
        {|}    jCenter_zX := Param_jCenter_zX;  jCenter_zY    := Param_jCenter_zY   ;
        {|}    ld_Zoom    := Param_ld_Zoom   ;
        {|}    Iteration  := Param_Iteration ;  Alt_Iteration := Param_Alt_Iteration;
        {|}    Color      := Param_Color     ;  Alt_Color     := Param_Alt_Color    ;  Offset   := Param_Offset  ;
        {|}    Automatik  := Param_Automatik ;  FPU_Mode      := Param_FPU_Mode     ;  CPU_Prev := Param_CPU_Prev;
        {|}  end;
        {|}
        {|}  FillRect (WindowDC, Area, Dark_Brush);                                 {Bildschirm lschen und alles neu berechnen}
        {|}  Set_N_Ctrl (Work); Estimate := NOP; Set_Variables_SYS; Kachel := 0;
        {|}
      end; {endif} {"Kachel = -1"}

      if (Message <> 0) then begin Show_Message; Kachel := 9999 end; {Override bei "Message", zudem wird der "Message_Timer" gestartet}

      if (Kachel <> 0) then begin Kachel := 9999; {+++} goto KACHEL_9999 {+++} end; {wirkt auch als Sicherheits-Funktion}

      FillDWord (Cursor, HD_Size, $808080); {"Cursor" fr die Fortschritts-Anzeige vorbereiten - bei erster Rekursion evtl. berflssig}

      Trigger_Paint_Action; {+++} goto READY {+++}; {evtl. roten Rahmen zeichnen, der Zyklus startet im nchsten Durchlauf}

    end; {endif} {"Kachel < 0"}

    {ZYKLUS: ab hier gilt sicher "0 <= Kachel <= 1023" - siehe Abfrage am Anfang}

    L := Thread_L [1, Kachel]; B := (Thread_R [Anzahl_Threads, Kachel] - L + 1);
    O := Thread_O [1, Kachel]; H := (Thread_U [Anzahl_Threads, Kachel] - O + 1); U := Window_Y - (O + H); {weil die DIB "verkehrt" herum ist}

    Break_Flag := false; Break_Cnt := HD_Size; Break_Anz := round (1.0e+07 / Iteration);
    if Kachel_Ausgabe then begin Break_Cnt := Break_Anz; QueryPerformanceCounter (QPC); Break_QPC := (QPC + QPC_pro_s) end;

    {+++} Start_Threads {+++};

    if Break_Flag then begin
      SetDIBitsToDevice (WindowDC, L, O, B, H, L, U, 0, Window_Y, @Cursor, dib, dib_RGB_Colors); Trigger_Paint_Action; {+++} exit {+++};
    end;

    if Kachel_Ausgabe then begin
      SetDIBitsToDevice (WindowDC, L, O, B, H, L, U, 0, Window_Y, @Pixel, dib, dib_RGB_Colors); {+++} goto NEXT {+++};
    end;

    QueryPerformanceCounter (QPC); if (QPC < Idle_QPC) then {+++} goto NEXT {+++}; {ab hier also KEINE "Kachel_Ausgabe"}

    with Area_z do begin left := L; right := (L + B); top := O; bottom := (O + H) end; {Bildschirm lschen und Cursor ausgeben}
    FillRect (WindowDC, Area, Dark_Brush); FillRect (WindowDC, Area_z, Light_Brush);

    NEXT: inc (Kachel);

    if (Kachel = 1024) then begin
      if (Rekursion = 6) then Rekursion :=  0; {Reihenfolge mu so sein!}
      if (Rekursion = 5) then Rekursion :=  1;
      if (Rekursion = 4) then Rekursion_Typ_4;                   {"Rekursion_Typ_4/3/2" setzt "Rekursion = 6/5", wordurch ein neuer Zyklus ...}
      if (Rekursion = 3) then Rekursion_Typ_3;                   {... gestartet wird, der am Ende dann in diese Abfrage-Kette hier hineinluft}
      if (Rekursion = 2) then Rekursion_Typ_2;
      if (Rekursion = 1) then Rekursion_Typ_1;
      if (Rekursion = 0) then goto ZYKLUSENDE;
      FillDWord (Cursor, HD_Size, $808080); Kachel := 0; {"Cursor" wieder lschen, im Folgenden wird der Zyklus neu gestartet}
    end;

    if Kachel_Ausgabe then begin
      Area_z.left := Thread_L [1, Kachel]; Area_z.right  := (Thread_R [Anzahl_Threads, Kachel] + 1); {Cursor ausgeben, auch fr "Kachel = 0"}
      Area_z.top  := Thread_O [1, Kachel]; Area_z.bottom := (Thread_U [Anzahl_Threads, Kachel] + 1);
      FillRect (WindowDC, Area_z, Light_Brush);
    end;

    Trigger_Paint_Action; {+++} goto READY {+++};       {weiter im Zyklus}

    ZYKLUSENDE: Kachel := 9999;                         {+++ Zyklus beendet +++}

    Laufzeit := (t - t_Start);                          {Laufzeit      }
    if Info_Screen then Show_Info_Screen;               {"Info_Screen" }
    if Wait_for_Init_Buf then Init_Buf else Write_Buf;  {Buffer        }
    if (Param_loaded <> 0) then Show_Message;           {Parameter-File, der "Message_Timer" wird gestartet und "Param_loaded" gelscht}

    KACHEL_9999: {ab hier gilt sicher "Kachel = 9999"}

    if Focus_Leave_Request then begin Focus_Leave_Request := false; Hide_Window; {+++} exit {+++} end; {Fenster ausblenden}

    if ((Call_Kachel <> -9) and (Call_Kachel <> -1)) or (not Fullscreen) or Message_Timer then {+++} goto READY {+++};

    {|}  L := Pixel [(Window_X - 4) + ((Window_Y - 4) * Window_X)];    {----------------------------------------------------------------------}
    {|}  H := L; byte32 (H) [2] := byte32 (L) [0];                     { XP blendet bei "Focus_Restore" oder Programm-Start die Taskleiste    }
    {|}          byte32 (H) [0] := byte32 (L) [2];                     { nicht immer aus. Deswegen wird ein Pixel vom unteren Bildschirm-Rand }
    {|}  L := GetPixel (WindowDC, Window_X - 4, Window_Y - 4);         { geprft und die erneute Bild-Ausgabe ber einen Timer angestoen.    }
    {|}                                                                {----------------------------------------------------------------------}
    {|}  if (L <> H) or (longword (L) = clr_Invalid) then begin
    {|}    if (XP_Try_Restore = -1) then begin XP_Try_Restore := 5 {Versuche}; SetTimer (Window, {No} 3, 500 {ms}, NIL) end;
    {|}  end else begin
    {|}    XP_Try_Restore := -1;   {wenn die Pixel immer ungleich wren, lscht sich der Timer ... }
    {|}  end;                      {... bei "XP_Try_Restore = 0" selbst und wird nie mehr gestartet}

    READY: {"Kachel = [0 .. 1023] oder 9999"}

    if Red_Frame then begin
      FillRect (WindowDC, Area_L, Red_Brush); FillRect (WindowDC, Area_O, Red_Brush); {roten Rahmen zeichnen}
      FillRect (WindowDC, Area_R, Red_Brush); FillRect (WindowDC, Area_U, Red_Brush);
    end;

  end; {endproc} {"Paint_Action"}


  procedure XP_Timer_Action; begin
    if (Kachel = 9999) and ((Call_Kachel = -9) or (Call_Kachel = -1)) and (XP_Try_Restore > 0) and (not Focus_lost) then begin
      dec (XP_Try_Restore); Kachel_Job (-9);
    end else begin
      KillTimer (Window, {No} 3);  {insbesondere bei "XP_Try_Restore <= 0" lscht sich der Timer selbst}
    end;
  end;

  procedure Message_Timer_Action; begin
    KillTimer (Window, {No} 4); Message_Timer := false; if (Kachel = 9999) and (Message < 0) and (not Focus_lost) then Kachel_Job (-9);
  end;


  {--- Tastatur-Eingabe ---}   {--------------------------------------------------------------------------------------------------------------}


  procedure CPU_Threshold (delta : longint); var M : byte; Msg : shortint; label NOP, NEW; begin

    CPU_Prev := (CPU_Prev + delta); Msg := 0;
    if (CPU_Prev < 0) then begin CPU_Prev := 0; Msg := 93; goto NOP end;
    if (CPU_Prev > 8) then begin CPU_Prev := 8; Msg := 94; goto NOP end;

    M := New_Mode;  if (M <> CPU_Mode) then goto NEW;  {$ifdef x32} if (delta = 0) and ((M = 1) or (M = 2)) then goto NEW; {$endif}

    NOP: if (Kachel = 9999) then begin Message := Msg; Kachel_Job (-9) end;  {... und ...}  {+++} exit {+++};

    NEW: Kachel_Job (-10);

  end; {endproc}


  {$define FPU_Threshold := FPU_Mode := (not FPU_Mode); CPU_Threshold (0) }


  procedure Toggle_Automatik; begin

    Automatik := (not Automatik); Buf_Automatik [Buf_Pointer] := Automatik; Estimate := NOP;         {"Buf_Automatik" und ebenso ...}
                                                                                                     {... "Buf_Alt_Iteration" merken}
    if Automatik then begin
      Alt_Iteration := Iteration; Alt_Color := Color; Offset := 0; CPU_Prev := 0;  {Automatik ON}
      Buf_Alt_Iteration [Buf_Pointer] := Buf_Iteration [Buf_Pointer];
    end else begin
      Alt_Iteration := 0; Alt_Color := 0; Buf_Alt_Iteration [Buf_Pointer] := 0;  {Automatik OFF}
    end;

  end; {endproc}


  procedure Zoom_in__Kachel_Job (I : longint); begin
    if (ld_Zoom >= 112) and (Kachel <> 9999) then {+++} exit {+++} else Kachel_Job (I);
  end;

  procedure Zoom_out_Kachel_Job (I : longint); begin
    if (ld_Zoom <= -4) and (Kachel <> 9999) then {+++} exit {+++} else Kachel_Job (I);
  end;

  procedure Toggle_Info_Screen; begin Info_Screen := (not Info_Screen) end;

  procedure Reset_Color; begin Color := 1.0; Alt_Color := Color; Offset := 0 end;  {"Color = 1.0" beliebig, weil Farb-Anpassung auf ON}

  procedure Shift_Spektrum (delta : longint); begin Offset := ((Offset + delta) and $FF) end;


  procedure KeyDown_Action (Key_Code : longword); var t_now : int64; begin

    if (Key_Code = $10) then begin Double_Key := (Double_Key or $100); {+++} exit {+++} end;   { Shift       merken                    }
    if (Key_Code = $11) then begin Double_Key := (Double_Key or $200); {+++} exit {+++} end;   { Ctrl        merken                    }
    if (Key_Code = $12) then begin Double_Key := 0;                    {+++} exit {+++} end;   { Sonderfall  ALT GR sendet $11 und $12 }

    if Mouse_pressed then {+++} exit {+++}; {beim "Mouse_Move" soll kein anderer "Kachel_Job" dazwischenfunken}

    t_now := t; if (t_now - t_last) < 250 {ms} then {+++} exit {+++}; t_last := t_now; {zu schnelle Eingaben verhindern}

    Key_Code := (Key_Code or Double_Key);

    case Key_Code of $09 : begin                              Kachel_Job (-10) end;   { TAB           Bild neu berechnen  }
           $35, $65, $74 : begin CPU_Threshold (+4);             {evtl.} {-10} end;   { 5, F5         CPU-Schwelle frher }
           $36, $66, $75 : begin CPU_Threshold (-4);             {evtl.} {-10} end;   { 6, F6         CPU-Schwelle spter }
                     $46 : begin FPU_Threshold { 0};             {evtl.} {-10} end;   { F             FPU- oder CPU-Mode  }
                     $08 : begin                     Read_Buf_Kachel_Job { -9} end;   { Backspace     ein Bild zurck     }
           $6B, $BB, $21 : begin                     Zoom_in__Kachel_Job ( -7) end;   { +, Bild Up    2-facher Zoom       }
           $6D, $BD, $22 : begin                     Zoom_out_Kachel_Job ( -6) end;   { -, Bild Down  Zoom out            }
                     $2E : begin Param_loaded := 0;           Kachel_Job ( -1) end;   { ENTF          System-Reset        }
               $243, $1B : begin DestroyWindow (Window);                       end;   { Ctrl-C, ESC   Programm beenden    }
    else begin

      if (Kachel <> 9999) then {+++} exit {+++}; {jetzt kommen die Funktionen, bei denen ein Zyklus beendet sein mu}

      Move_X := 0; Move_Y := 0; {"Move" vorbereiten}

      case Key_Code of $0D : begin                              Kachel_Job (-9) end;   { ENTER              Bild ausgeben            }
        $20, $31, $61, $70 : begin Toggle_Info_Screen;          Kachel_Job (-9) end;   { 1, F1, Space       "Info_Screen"            }
             $32, $62, $71 : begin Toggle_Automatik;            Kachel_Job (-9) end;   { 2, F2              Automatik                }
             $33, $63, $72 : begin                              Kachel_Job (-4) end;   { 3, F3              weniger Iterationen      }
             $34, $64, $73 : begin                              Kachel_Job (-5) end;   { 4, F4              mehr Iterationen         }
                       $4E : begin Color := (Color / 1.1);      Kachel_Job (-9) end;   { N                  weniger Farbe            }
                       $4D : begin Color := (Color * 1.1);      Kachel_Job (-9) end;   { M                  mehr Farbe               }
                      $24E : begin Shift_Spektrum (-16);        Kachel_Job (-9) end;   { Ctrl-N             Spektrum links schieben  }
                      $24D : begin Shift_Spektrum (+16);        Kachel_Job (-9) end;   { Ctrl-M             Spektrum rechts schieben }
           $49, $149, $249 : begin Message := 3;                Kachel_Job (-9) end;   { I, Ctrl/Shift-I    "created by" einblenden  }
                       $25 : begin Move_X := +(Window_X shr 2); Kachel_Job (-2) end;   { Pfeil L            Bild verschieben         }
                       $27 : begin Move_X := -(Window_X shr 2); Kachel_Job (-2) end;   { Pfeil R            dto.                     }
                       $26 : begin Move_Y := +(Window_Y shr 2); Kachel_Job (-2) end;   { Pfeil O            dto.                     }
                       $28 : begin Move_Y := -(Window_Y shr 2); Kachel_Job (-2) end;   { Pfeil U            dto.                     }
       $253, $39, $69, $78 : begin Write_Files;                 Kachel_Job (-9) end;   { 9, F9, Ctrl-S      Bild speichern           }
             $5B, $5C, $7A : begin Hide_Window;                                 end;   { WINDOWS-Keys, F11  Fenster ausblenden       }
      end; {endcase}

    end; {endelse} end; {endcase}

  end; {endproc}


  procedure KeyUp_Action (Key_Code : longword); begin
    if (Key_Code = $10) then Double_Key := (Double_Key and $EFF);   { Shift losgelassen merken }
    if (Key_Code = $11) then Double_Key := (Double_Key and $DFF);   { Ctrl  losgelassen merken }
  end; {endproc}


  {--- Maus-Verwaltung ---}   {---------------------------------------------------------------------------------------------------------------}


  function Mouse (z : Field) : boolean; begin Mouse := (z.L <= Mouse_X) and (Mouse_X <= z.R) and (z.O <= Mouse_Y) and (Mouse_Y <= z.U) end;


  procedure LButtonDown_Action (X_Pos, Y_Pos : word); var t_now : int64; begin

    t_now := t; if (t_now - t_last) < 250 {ms} then {+++} exit {+++}; t_last := t_now;  {zu schnelle Eingaben verhindern}

    if Mouse_pressed or Focus_lost then {+++} exit {+++};  {nicht berreagieren, z.B. kann der Maus-Timer noch kommen}

    if (not Fullscreen) and (not Mouse_Timer) then begin SetTimer (Window, {No} 2, 500 {ms}, NIL); Mouse_Timer := true end;  {Maus-Timer ON}

    Mouse_pressed := true; Move_happens := false;

    Mouse_X := X_Pos; Mouse_Y := Y_Pos; Move_X := 0; Move_Y := 0;  {... und ...}  Mouse_Color := Color;

  end; {endproc}


  procedure Mouse_Move_Action (X_Pos, Y_Pos : word); var Shift : double; begin

    if (not Mouse_pressed) or Focus_lost or (Kachel <> 9999) then {+++} exit {+++};  {fr alle "Moves" mu der Zyklus beendet sein}

    Move_X := (X_Pos - Mouse_X); Move_Y := (Y_Pos - Mouse_Y);

    if (Move_X = 0) and (Move_Y = 0) then {+++} exit {+++};

    Move_happens := true; {Meldung an den Maus-Timer}

    if Mouse (Field_Color) then begin
      Shift := exp (5.0 * (abs (Move_X) / Window_X));
      if (Move_X >= 0) then Color := (Mouse_Color * Shift) else Color := (Mouse_Color / Shift);
      Kachel_Job (-9); {Bild mit neuer Farbe ausgeben}
    end else begin
      Kachel_Job (-3); {Bild nur oberflchlich auf dem Bildschirm verschieben}
    end;

  end; {endproc}


  procedure LButtonUp_Action (X_Pos, Y_Pos : word); begin

    if Mouse_Timer then begin KillTimer (Window, {No} 2); Mouse_Timer := false end;  {Maus-Timer abschalten}

    if Focus_lost then {+++} exit {+++};
                                                                                                {Sicherheits-Funktion - passiert ... }
    if (Kachel = -3) then begin Mouse_pressed := false; Kachel_Job (-2); {+++} exit {+++} end;  {... vor allem bei langsamen Rechnern}

    if (not Mouse_pressed) then {+++} exit {+++};  {... ansonsten ...}  Mouse_pressed := false;

    Move_X := (X_Pos - Mouse_X); Move_Y := (Y_Pos - Mouse_Y);

    if (Move_X = 0) and (Move_Y = 0) and (not Move_happens) then begin {normaler Mausklick}

      if (Kachel = 9999) and (Message < 0) then begin
        if Mouse (Field_Message) then begin Kachel_Job (-9); {+++} exit {+++} end;  { Message-Anzeige ausblenden }
      end;

      if Mouse (Field_Back    ) then begin                         Read_Buf_Kachel_Job {-9}; {+++} exit {+++} end;   { Bild zurck      }
      if Mouse (Field_Reset   ) then begin Param_loaded := 0;               Kachel_Job (-1); {+++} exit {+++} end;   { System-Reset     }
      if Mouse (Field_Zoom_x16) then begin                         Zoom_in__Kachel_Job (-8); {+++} exit {+++} end;   { 16-facher Zoom   }
      if Mouse (Field_Zoom_x2 ) then begin                         Zoom_in__Kachel_Job (-7); {+++} exit {+++} end;   { 2-facher Zoom    }
      if Mouse (Field_Zoom_out) then begin                         Zoom_out_Kachel_Job (-6); {+++} exit {+++} end;   { Zoom out         }
      if Mouse (Field_Exit    ) then begin DestroyWindow (Window);                           {+++} exit {+++} end;   { Programm beenden }

      if (Kachel <> 9999) then {+++} exit {+++};  {jetzt kommen die Funktionen, bei denen ein Zyklus beendet sein mu}

      if Info_Screen then begin
        if Mouse (Field_Info_Auto) then begin Toggle_Automatik;   Kachel_Job ( -9); {+++} exit {+++} end;   { Automatik           }
        if Mouse (Field_Info_Less) then begin                     Kachel_Job ( -4); {+++} exit {+++} end;   { weniger Iterationen }
        if Mouse (Field_Info_More) then begin                     Kachel_Job ( -5); {+++} exit {+++} end;   { mehr Iterationen    }
        if Mouse (Field_Info_Prev) then begin CPU_Threshold (+4);    {evtl.} {-10}; {+++} exit {+++} end;   { CPU-Schwelle frher }
        if Mouse (Field_Info_Late) then begin CPU_Threshold (-4);    {evtl.} {-10}; {+++} exit {+++} end;   { CPU-Schwelle spter }
        if Mouse (Field_Info_Save) then begin Write_Files;        Kachel_Job ( -9); {+++} exit {+++} end;   { Bild speichern      }
        if Mouse (Field_Info_F11 ) then begin Hide_Window;                          {+++} exit {+++} end;   { Fenster ausblenden  }
      end;

      if Mouse (Field_Info ) then begin Toggle_Info_Screen; Kachel_Job (-9); {+++} exit {+++} end;   { "Info_Screen"      }
      if Mouse (Field_Color) then begin Reset_Color;        Kachel_Job (-9); {+++} exit {+++} end;   { Farbe zurcksetzen }

    end else begin {ein "Move" ist erfolgt}

      if (Kachel <> 9999) then {+++} exit {+++};  {fr alle "Moves" mu der Zyklus beendet sein}

      if Mouse (Field_Color) then Kachel_Job (-9) else Kachel_Job (-2);  {Farbe ndern oder Bild im Speicher verschieben}

    end; {endelse} {"Move"}

  end; {endproc}
                                                                                                {warten, bis kein "Move" mehr stattfindet, ...}
                                                                                                {... dann "ButtonUp" simulieren - oder ein ...}
  procedure Mouse_Timer_Action; begin {ersetzt "TrackMouseEvent"}                               {... echter "ButtonUp" den Timer ausschaltet  }

    if Move_happens and (not Focus_lost) then begin Move_happens := false; {+++} exit {+++} end;

    LButtonUp_Action ((Mouse_X + Move_X), (Mouse_Y + Move_Y)); {"ButtonUp" mit der letzten bekannten Maus-Position simulieren}

  end; {endproc}


  procedure Mouse_Wheel_Action (Direction : word); begin
    if (Kachel = 9999) and (not Focus_lost) then begin
      if ((Direction and $8000) = 0) then Color := (Color * 1.02) else Color := (Color / 1.02); Kachel_Job (-9); {Farbe ndern}
    end;
  end;


  procedure RButtonUp_Action (X_Pos, Y_Pos : word); begin
    if (Kachel = 9999) and (not Focus_lost) then begin Info_Screen := (not Info_Screen); Kachel_Job (-9) end; {"Info_Screen"}
  end;


  {--- Programm-Start ---}   {----------------------------------------------------------------------------------------------------------------}


  procedure Init_System; begin {Ablauf-Steuerung initialisieren}

    Wait_for_first_Paint := true; Wait_for_Init_Buf := true; {erstmalige Ablufe abwarten}

    Double_Key := 0; Mouse_pressed := false; Mouse_Timer := false; {Status von Tastatur und Maus}

    Info_Screen := true; Message_Timer := false; Message := 0; {Status der Anzeige}

    Focus_lost := false; Focus_Restore_locked := false; Focus_Leave_Request := false; XP_Try_Restore := -1; {zum Fenster ein-/ausblenden}

    QueryPerformanceFrequency (QPC_pro_s); t_last := 0; {Zeitmessung initialisieren}

    FillDWord (N_Anf, N_real_Size, 0); {Rnder vom "N_array" auf "N = 0" - schnell, kann aber unsaubere Rnder verursachen}

    Kachel := -1; {in der "Paint_Action" wird der Rest initialisiert - einschlielich "User_Create" statt "wm_Create"}

  end; {endproc}


  {|} {--------------------------------------------------------------------------------------------------------} {|}
  {|}                                                                                                            {|}
  {|}   {$include MBEx-API.hdr}   {API-Kern und Hauptprogramm}                                                   {|}
  {|}                                                                                                            {|}
  {|}     { Read_ParamStr;             }                                                                         {|}
  {|}     { Get_Screen_Metrics;        }                                                                         {|}
  {|}     { Set_Window_Size;           }                                                                         {|}
  {|}     { Open_Window;               }                                                                         {|}
  {|}     { Make_Threads;              }                                                                         {|}
  {|}     { Split_Window_into_Kacheln; }                                                                         {|}
  {|}     { Make_Spektrum;             }                                                                         {|}
  {|}     { Make_Mouse_Buttons;        }                                                                         {|}
  {|}     { Check_RAM;                 }                                                                         {|}
  {|}     { Init_System;               }                                                                         {|}
  {|}     { Enter_Message_Loop;        }                                                                         {|}
  {|}                                                                                                            {|}
  {|} {--------------------------------------------------------------------------------------------------------} {|}


end. {endprog}


{--- Ende der Datei ---}   {------------------------------------------------------------------------------------------------------------------}

